#!/usr/bin/env python3

import argparse
import os
from tempfile import NamedTemporaryFile, mkdtemp
from shutil import rmtree

import multiprocessing
from functools import partial

import numpy as np
from scipy import stats

import pysam
import pysamstats
import pybedtools


def get_bamcoverage(bamfile, fafile, chrom, window_size=20):
    mybam = pysam.AlignmentFile(bamfile)
    a = pysamstats.load_binned("coverage", mybam,
                               fafile=fafile,
                               chrom=chrom,
                               window_size=window_size)
    return a


def get_coverage_cutoff(coverage):
    counts = coverage.reads_all[np.where(coverage.reads_all)]

    mode = stats.mode(counts)
    stdev = np.std(counts)
    # recommanded by SpeedSeq
    return mode[0] + 3 * stdev


def dump_to_bed(overcovered_regions, fout, window_size):
    for chrom, pos in zip(overcovered_regions.chrom,
                          overcovered_regions.pos):
        fout.write("%s\t%d\t%d\n" % (chrom.decode("utf-8"),
                                     pos - int(window_size/2),
                                     pos + int(window_size/2)))


def detect_high_coverage(chrom, bamfile, fastafile, tempdir,
                         window_size=20, cutoff=None):
    print("Detecting high coverage regions in %s for %s" % (bamfile, chrom))
    with NamedTemporaryFile(dir=tempdir, mode="w", delete=False) as output:
        coverage = get_bamcoverage(bamfile, fastafile, chrom, window_size)
        if cutoff is None:
            cutoff = get_coverage_cutoff(coverage)
        overcovered_regions = coverage[np.where(coverage.reads_all > cutoff)]
        dump_to_bed(overcovered_regions, output, window_size)
    return output.name, cutoff


def parallel_runs(chromosomes, bamfile, fastafile, tempdir,
                  window_size, cutoff, cores=1):
    pool = multiprocessing.Pool(processes=cores)
    launch_chrom = partial(detect_high_coverage,
                           bamfile=bamfile,
                           fastafile=fastafile,
                           tempdir=tempdir,
                           window_size=window_size,
                           cutoff=cutoff)  # only remaining param is chrom
    result_list = pool.map(launch_chrom, chromosomes)
    return [r[0] for r in result_list]


def main(args):
    bamfile = args.bam
    fastafile = args.fasta
    chroms = args.chroms
    output = args.output
    window_size = args.window_size
    threads = args.threads
    chromosomes = chroms.split(',')

    tempdir = mkdtemp(dir=".")

    outputfiles = []
    # pop fist chromosome
    chrom = chromosomes.pop(0)
    outputfile, cutoff = detect_high_coverage(chrom, bamfile, fastafile,
                                              tempdir, window_size=window_size)
    outputfiles.append(outputfile)

    outfiles = parallel_runs(chromosomes, bamfile, fastafile,
                             tempdir, window_size, cutoff,
                             cores=threads)
    outputfiles.extend(outfiles)

    with NamedTemporaryFile(dir=tempdir, mode="w", delete=False) as bedout:
        for fname in outputfiles:
            with open(fname) as infile:
                for line in infile:
                    bedout.write(line)
            os.remove(fname)
    bed = pybedtools.BedTool(bedout.name)
    merged_bed = bed.merge()
    os.remove(bedout.name)
    merged_bed.saveas(output)

    rmtree(tempdir)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Extract entries for a "
                                     "specific chromosome from a bam file")
    parser.add_argument('-b', '--bam', required=True,
                        help='the bam file')
    parser.add_argument('-f', '--fasta', required=True,
                        help='the reference genome')
    parser.add_argument('-c', '--chroms', required=True,
                        help='comma separated list of chromosomes')
    parser.add_argument('-o', '--output', required=True,
                        help='output file')
    parser.add_argument('-w', '--window_size', type=int, default=20,
                        help='the size of the window for coverage estimation')
    parser.add_argument('-t', '--threads',  type=int, default=1,
                        help='number of threads')
    args = parser.parse_args()
    main(args)
